#!usr/bin/env/python3
# -*- coding: utf-8 -*-

import os
import sys
import click
import stat
import shutil
import tarfile
import fileinput

from minitrino.cli import pass_environment
from minitrino import utils
from minitrino import errors as err
from minitrino.settings import SNAPSHOT_ROOT_FILES
from minitrino.settings import PROVISION_SNAPSHOT_TEMPLATE
from minitrino.settings import LIB
from minitrino.settings import MODULE_ROOT
from minitrino.settings import MODULE_ADMIN
from minitrino.settings import MODULE_CATALOG
from minitrino.settings import MODULE_SECURITY
from minitrino.settings import MODULE_RESOURCES
from minitrino.settings import SCRUB_KEYS


@click.command(
    "snapshot",
    help=(
        """Create a snapshot of a Minitrino environment. A tarball is placed in
        the Minitrino `lib/snapshots/` directory.

        To take a snapshot of an active environment, leave the `--module` and
        option out of the command. 

        To take a snapshot of modules whether they are active or not, specify
        the modules via the `--module` option."""
    ),
)
@click.option(
    "-m",
    "--module",
    "modules",
    default=[],
    type=str,
    multiple=True,
    help=("""A specific module to snapshot."""),
)
@click.option(
    "-n",
    "--name",
    required=True,
    type=str,
    help=(
        """Basename of the resulting snapshot tarball file. Allowed characters:
        alphanumerics, hyphens, and underscores."""
    ),
)
@click.option(
    "-d",
    "--directory",
    type=click.Path(),
    help=(
        """Directory to save the resulting snapshot file in. Defaults to the
        snapshots directory in the Minitrino library."""
    ),
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help=("""Overwrite the file if it already exists."""),
)
@click.option(
    "--no-scrub",
    is_flag=True,
    default=False,
    help=(
        """Do not scrub sensitive data from user config file.

        WARNING: all sensitive information (passwords and keys) will be kept in
        the user config file added to the snapshot. Only use this if you are
        prepared to share those secrets with another person."""
    ),
)
@utils.exception_handler
@pass_environment
def cli(ctx, modules, name, directory, force, no_scrub):
    """Snapshot command for Minitrino."""

    # The snapshot temp files are saved in ~/.minitrino/snapshots/<name>
    # regardless of the directory provided. The artifact (tarball) will go
    # to either the default directory or the user-provided directory.

    utils.check_lib(ctx)
    modules = list(modules)

    if directory and not os.path.isdir(directory):
        raise err.UserError(
            f"Cannot save snapshot in nonexistent directory: {directory}",
            "Pick any directory that exists, and this will work.",
        )

    if not directory:
        directory = os.path.join(ctx.minitrino_lib_dir, "snapshots")

    validate_name(name)
    check_exists(name, directory, force)

    if modules:
        ctx.logger.info(f"Creating snapshot of specified modules...")
        modules = utils.check_dependent_modules(ctx, modules)
        snapshot_runner(name, no_scrub, False, modules, directory)
    else:
        modules = ctx.modules.get_running_modules()
        if not modules:
            ctx.logger.verbose(
                f"No running Minitrino modules to snapshot. Snapshotting "
                f"Trino resources only.",
            )
        else:
            ctx.logger.info(f"Creating snapshot of active environment...")
            modules = utils.check_dependent_modules(ctx, modules)
        snapshot_runner(name, no_scrub, True, modules, directory)

    check_complete(name, directory)
    ctx.logger.info(
        f"Snapshot complete and saved at path: {os.path.join(directory, name)}.tar.gz"
    )


@pass_environment
def validate_name(ctx, name):
    """Validates the chosen filename for correct input."""

    for char in name:
        if all((char != "_", char != "-", not char.isalnum())):
            raise err.UserError(
                f"Illegal character found in provided filename: '{char}'. ",
                f"Alphanumeric, hyphens, and underscores are allowed. "
                f"Rename and retry.",
            )


@pass_environment
def check_exists(ctx, name, directory, force):
    """Checks if the resulting tarball exists. If it exists, the user is
    prompted to overwrite the existing file."""

    if force:
        return

    snapshot_file = os.path.abspath(os.path.join(directory, f"{name}.tar.gz"))
    if os.path.isfile(snapshot_file):
        response = ctx.logger.prompt_msg(
            f"Snapshot file {name}.tar.gz already exists. Overwrite? [Y/N]"
        )
        if not utils.validate_yes(response):
            ctx.logger.info(f"Opted to skip snapshot.")
            sys.exit(0)


@pass_environment
def prepare_snapshot_dir(ctx, name, active, no_scrub, modules):
    """Checks if the snapshot temp directory exists. If it does, clears
    files/directories inside of it. If it doesn't, (1) creates it and clones the
    library structure, (2) adds a Bash file that can be executed to spin up the
    environment as it was snapshotted.

    Returns the absolute path of the named snapshot directory."""

    if os.path.isdir(ctx.snapshot_dir):
        ctx.logger.verbose(
            "Snapshot temp directory exists. Removing and recreating...",
        )
        shutil.rmtree(ctx.snapshot_dir)
        os.mkdir(ctx.snapshot_dir)
    else:
        ctx.logger.verbose("Snapshot directory does not exist. Creating...")
        os.mkdir(ctx.snapshot_dir)

    snapshot_name_dir = clone_lib_dir(name)
    handle_copy_config_file(snapshot_name_dir, no_scrub)
    build_snapshot_command(snapshot_name_dir, modules, active)

    return snapshot_name_dir


@pass_environment
def build_snapshot_command(ctx, snapshot_name_dir, modules=[], active=True):
    """Builds a basic shell command that can be used to provision an environment
    with the minitrino CLI. Used for snapshot purposes."""

    command_string = build_command_string(modules)
    create_snapshot_command_file(command_string, snapshot_name_dir)


@pass_environment
def build_command_string(ctx, modules=[]):
    """Builds a command string that can be used to create an environment with
    the designated modules."""

    option_string = ""

    if modules:
        options = ""
        for module in modules:
            options += f"--module {module} "
        option_string = f"{options}"

    bash_source = '"${BASH_SOURCE%/*}"'
    command_string = (
        f"minitrino -v --env LIB_PATH={bash_source}/lib provision {option_string}\n\n"
    )

    return command_string.replace("  ", " ")


@pass_environment
def create_snapshot_command_file(ctx, command_string="", snapshot_name_dir=""):
    """Creates an .sh file in the minitrino directory for usage by the snapshot
    command. This way, a similar command used to provision the environment is
    preserved."""

    file_dest = os.path.join(snapshot_name_dir, "provision-snapshot.sh")
    ctx.logger.verbose(
        f"Creating snapshot command to file at path: {file_dest}",
    )

    # Create provisioning command snapshot file from template and make it
    # executable
    try:
        with open(file_dest, "w") as provision_command_file:
            provision_command_file.write(PROVISION_SNAPSHOT_TEMPLATE.lstrip())
        st = os.stat(file_dest)
        os.chmod(
            file_dest,
            st.st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH,
        )
    except Exception as e:
        utils.handle_exception(e, ctx.verbose)

    with open(file_dest, "a") as provision_snapshot_file:
        provision_snapshot_file.write(command_string)


@pass_environment
def clone_lib_dir(ctx, name):
    """Clones the library directory structure and necessary top-level files in
    preparation for copying over module directories.

    Returns the absolute path of the named snapshot directory."""

    snapshot_name_dir = os.path.join(ctx.snapshot_dir, name)
    os.makedirs(os.path.join(snapshot_name_dir, LIB, MODULE_ROOT, MODULE_ADMIN))
    os.mkdir(os.path.join(snapshot_name_dir, LIB, MODULE_ROOT, MODULE_CATALOG))
    os.mkdir(os.path.join(snapshot_name_dir, LIB, MODULE_ROOT, MODULE_SECURITY))
    os.mkdir(os.path.join(snapshot_name_dir, LIB, MODULE_ROOT, MODULE_RESOURCES))

    # Copy root lib files to snapshot
    for filename in os.listdir(ctx.minitrino_lib_dir):
        if filename in SNAPSHOT_ROOT_FILES:
            file_path = os.path.join(ctx.minitrino_lib_dir, filename)
            if os.path.isfile(file_path):
                shutil.copy(file_path, os.path.join(snapshot_name_dir, LIB))
            elif os.path.isdir(file_path):
                shutil.copytree(
                    file_path, os.path.join(snapshot_name_dir, LIB, filename)
                )

    # Copy everything from lib/modules/resources
    resources_dir = os.path.join(ctx.minitrino_lib_dir, MODULE_ROOT, MODULE_RESOURCES)
    for filename in os.listdir(resources_dir):
        file_path = os.path.join(resources_dir, filename)
        shutil.copy(
            file_path,
            os.path.join(snapshot_name_dir, LIB, MODULE_ROOT, MODULE_RESOURCES),
        )

    return snapshot_name_dir


@pass_environment
def handle_copy_config_file(ctx, snapshot_name_dir, no_scrub):
    """Handles the copying of the user config file to the named snapshot
    directory. Calls `scrub_config_file()` if `no_scrub` is True."""

    if no_scrub:
        response = ctx.logger.prompt_msg(
            f"All sensitive information in user config will be added to the snapshot. Continue? [Y/N]"
        )
        if utils.validate_yes(response):
            copy_config_file(snapshot_name_dir, no_scrub)
        else:
            ctx.logger.info(f"Opted to scrub sensitive user config data.")
            copy_config_file(snapshot_name_dir)
    else:
        copy_config_file(snapshot_name_dir)


@pass_environment
def copy_config_file(ctx, snapshot_name_dir, no_scrub=False):
    """Copies user config file to the named snapshot directory."""

    if os.path.isfile(ctx.config_file):
        shutil.copy(ctx.config_file, snapshot_name_dir)
    else:
        ctx.logger.warn(
            f"No user config file at path: {ctx.config_file}. Will not be added to snapshot.",
        )
        return

    if not no_scrub:
        scrub_config_file(snapshot_name_dir)


@pass_environment
def scrub_config_file(ctx, snapshot_name_dir):
    """Scrubs the user config file of sensitive data."""

    snapshot_config_file = os.path.join(snapshot_name_dir, "minitrino.cfg")
    if os.path.isfile(snapshot_config_file):
        for line in fileinput.input(snapshot_config_file, inplace=True):
            if "=" in line:
                print(line.replace(line, scrub_line(line)))
            else:
                print(line.replace(line, line.rstrip()))
    else:
        ctx.logger.warn(
            f"No user config file at path: {ctx.config_file}. Nothing to scrub.",
        )


@pass_environment
def scrub_line(ctx, line):
    """Scrubs a line from a snapshot config file. Returns the scrubbed line."""

    # If the key has a substring that matches any of the scrub keys, we know
    # it's an item whose value needs to be scrubbed
    line = utils.parse_key_value_pair(line, err_type=err.UserError)
    if any(item in line[0].lower() for item in SCRUB_KEYS):
        line[1] = "*" * 20

    return "=".join(line)


@pass_environment
def copy_module_dirs(ctx, snapshot_name_dir, modules=[]):
    """Copies module directories into the named snapshot directory."""

    for module in modules:
        module_dir = ctx.modules.data.get(module, {}).get("module_dir", "")
        module_type = ctx.modules.data.get(module, {}).get("type", "")
        dest_dir = os.path.join(
            os.path.join(snapshot_name_dir, LIB, MODULE_ROOT, module_type),
            os.path.basename(module_dir),
        )
        shutil.copytree(module_dir, dest_dir)


@pass_environment
def create_named_tarball(ctx, name, snapshot_name_dir, save_dir):
    """Creates a tarball of the named snapshot directory and placed in the
    library's snapshot directory."""

    with tarfile.open(os.path.join(save_dir, f"{name}.tar.gz"), "w:gz") as tar:
        tar.add(snapshot_name_dir, arcname=os.path.basename(snapshot_name_dir))


def snapshot_runner(name, no_scrub, active, modules=[], directory=""):
    """Executes sequential snapshot command functions."""

    snapshot_name_dir = prepare_snapshot_dir(name, active, no_scrub, modules)
    copy_module_dirs(snapshot_name_dir, modules)
    create_named_tarball(name, snapshot_name_dir, directory)


@pass_environment
def check_complete(ctx, name, directory):
    """Checks if the snapshot completed. If detected as incomplete, exists with
    a non-zero status code."""

    snapshot_file = os.path.join(directory, f"{name}.tar.gz")
    if not os.path.isfile(snapshot_file):
        raise err.MinitrinoError(f"Snapshot tarball failed to write to {snapshot_file}")



#!usr/bin/env/python3
# -*- coding: utf-8 -*-

# Instead of using the Docker SDK, this script will form a Docker Compose
# command string to execute straight through the docker compose CLI. This is
# required because the Docker SDK does not support communication with Compose
# files, and Minitrino benefits hugely from Docker Compose.

import os
import re
import stat
import hashlib
import time
import click
import yaml

from minitrino.cli import pass_environment
from minitrino import utils
from minitrino import errors as err
from minitrino.settings import RESOURCE_LABEL
from minitrino.settings import ETC_TRINO
from minitrino.settings import LIC_VOLUME_MOUNT
from minitrino.settings import LIC_MOUNT_PATH
from minitrino.settings import DUMMY_LIC_MOUNT_PATH
from minitrino.settings import TRINO_CONFIG
from minitrino.settings import TRINO_JVM_CONFIG
from minitrino.settings import WORKER_CONFIG_PROPS

from docker.errors import NotFound
from concurrent.futures import ThreadPoolExecutor, as_completed


@click.command(
    "provision",
    help=(
        """Provision an environment based on specified modules. All options are
        optional and can be left empty."""
    ),
)
@click.option(
    "-m",
    "--module",
    "modules",
    default=[],
    type=str,
    multiple=True,
    help=("""A specific module to provision."""),
)
@click.option(
    "-w",
    "--workers",
    "workers",
    default=0,
    type=int,
    help=("""The number of workers to provision (default: 0)."""),
)
@click.option(
    "-n",
    "--no-rollback",
    is_flag=True,
    default=False,
    help=(
        """Do not rollback provisioned resources in the event of an
        error."""
    ),
)
@click.option(
    "-d",
    "--docker-native",
    default="",
    type=str,
    help=(
        """Appends native docker compose commands to the generated
        docker compose shell command. Run `docker compose up --help` to see all
        available options.

        Example: minitrino provision --docker-native --build

        Example: minitrino provision --docker-native '--remove-orphans
        --force-recreate'"""
    ),
)
@utils.exception_handler
@pass_environment
def cli(ctx, modules, workers, no_rollback, docker_native):
    """Provision command for Minitrino. If the resulting docker compose command
    is unsuccessful, the function exits with a non-zero status code."""

    utils.check_daemon(ctx.docker_client)
    utils.check_lib(ctx)
    utils.check_starburst_ver(ctx)
    modules = list(modules)
    modules = append_running_modules(modules)
    modules = utils.check_dependent_modules(ctx, modules)
    check_compatibility(modules)
    check_enterprise(modules)
    check_volumes(modules)

    if not modules:
        ctx.logger.info(
            f"No modules specified. Provisioning standalone Trino container..."
        )
    else:
        for module in modules:
            if not ctx.modules.data.get(module, False):
                raise err.UserError(
                    f"Invalid module: '{module}'. It was not found "
                    f"in the Minitrino library at {ctx.minitrino_lib_dir}"
                )

    try:
        cmd_chunk = chunk(modules)
        compose_cmd = build_command(docker_native, cmd_chunk)

        ctx.cmd_executor.execute_commands(compose_cmd, environment=ctx.env)
        initialize_containers()

        c_restart = execute_bootstraps(modules)
        c_restart = write_trino_cfg(c_restart, modules)
        check_dup_cfgs()
        c_restart = provision_workers(c_restart, workers)
        restart_containers(c_restart)
        ctx.logger.info(f"Environment provisioning complete.")

    except Exception as e:
        rollback(no_rollback)
        utils.handle_exception(e)


@pass_environment
def append_running_modules(ctx, modules=[]):
    """Checks if any modules are already running. If they are, they are appended
    to the provided modules list and the updated list is returned."""

    ctx.logger.verbose("Checking for running modules...")
    running_modules = ctx.modules.get_running_modules()

    if running_modules:
        ctx.logger.verbose(
            f"Identified the following running modules: {running_modules}. "
            f"Appending the running module list to the list of modules to provision.",
        )

    modules = list(modules)
    modules.extend(running_modules)
    return list(set(modules))


@pass_environment
def check_compatibility(ctx, modules=[]):
    """Checks if any of the provided modules are mutually exclusive of each
    other. If they are, a user error is raised."""

    for module in modules:
        incompatible = ctx.modules.data.get(module, {}).get("incompatibleModules", [])
        if not incompatible:
            continue
        for module_inner in modules:
            if (module_inner in incompatible) or (
                incompatible[0] == "*" and len(modules) > 1
            ):
                raise err.UserError(
                    f"Incompatible modules detected. Tried to provision module "
                    f"'{module_inner}', but found that the module is incompatible "
                    f"with module '{module}'. Incompatible modules listed for module "
                    f"'{module}' are: {incompatible}",
                    f"You can see which modules are incompatible with this module by "
                    f"running 'minitrino modules -m {module}'",
                )


@pass_environment
def check_enterprise(ctx, modules=[]):
    """Checks if any of the provided modules are Starburst Enterprise features.
    If they are, we check that a pointer to a SEP license is provided."""

    ctx.logger.verbose(
        "Checking for SEP license for enterprise modules...",
    )

    yaml_path = os.path.join(ctx.minitrino_lib_dir, "docker-compose.yml")
    with open(yaml_path) as f:
        yaml_file = yaml.load(f, Loader=yaml.FullLoader)
    volumes = yaml_file.get("services", {}).get("trino", {}).get("volumes", [])

    if LIC_VOLUME_MOUNT not in volumes:
        raise err.UserError(
            f"The required license volume in the library's root docker-compose.yml "
            f"is either commented out or deleted: {yaml_path}. For reference, "
            f"the proper volume mount is: '{LIC_VOLUME_MOUNT}'"
        )

    enterprise_modules = []
    for module in modules:
        if ctx.modules.data.get(module, {}).get("enterprise", False):
            enterprise_modules.append(module)

    if enterprise_modules:
        if not ctx.env.get("LIC_PATH", False):
            raise err.UserError(
                f"Module(s) {enterprise_modules} requires a Starburst license. "
                f"You must provide a path to a Starburst license via the "
                f"LIC_PATH environment variable"
            )
        ctx.env.update({"LIC_MOUNT_PATH": LIC_MOUNT_PATH})
    elif ctx.env.get("LIC_PATH", False):
        ctx.env.update({"LIC_MOUNT_PATH": LIC_MOUNT_PATH})
    else:
        ctx.env.update({"LIC_PATH": "./modules/resources/dummy.license"})
        ctx.env.update({"LIC_MOUNT_PATH": DUMMY_LIC_MOUNT_PATH})


@pass_environment
def check_volumes(ctx, modules=[]):
    """Checks if any of the modules have persistent volumes and issues a warning
    to the user if so."""

    ctx.logger.verbose(
        "Checking modules for persistent volumes...",
    )

    for module in modules:
        if ctx.modules.data.get(module, {}).get("yaml_dict", {}).get("volumes", {}):
            ctx.logger.warn(
                f"Module '{module}' has persistent volumes associated with it. "
                f"To delete these volumes, remember to run `minitrino remove --volumes`.",
            )


@pass_environment
def chunk(ctx, modules=[]):
    """Builds docker compose command chunk for the chosen modules. Returns a
    command chunk string."""

    chunk = []
    for module in modules:
        yaml_file = ctx.modules.data.get(module, {}).get("yaml_file", "")
        chunk.extend(f"-f {yaml_file} \\\n")
    return "".join(chunk)


@pass_environment
def build_command(ctx, docker_native="", chunk=""):
    """Builds a formatted docker compose command for shell execution. Returns a
    docker compose command string."""

    cmd = []
    compose_env_string = ""
    for k, v in ctx.env.items():
        compose_env_string += f'{k.upper()}="{v}" '

    cmd.extend(
        [
            compose_env_string,
            "\\\n",
            "docker compose -f ",
            os.path.join(ctx.minitrino_lib_dir, "docker-compose.yml"),
            " \\\n",
            chunk,  # Module YAML paths
            "up -d --no-recreate",
        ]
    )

    if docker_native:
        ctx.logger.verbose(
            f"Received native Docker Compose options: '{docker_native}'",
        )
        cmd.extend([" ", docker_native])
    return "".join(cmd)


@pass_environment
def execute_bootstraps(ctx, modules=[]):
    """Executes bootstrap script for each container that has one––bootstrap
    scripts will only execute once the container is fully running to prevent
    conflicts with procedures executing as part of the container's entrypoint.
    After each script executes, the relevant container is added to a restart
    list.

    Returns a list of containers names which had bootstrap scripts executed
    inside of them."""

    services = []
    for module in modules:
        yaml_file = ctx.modules.data.get(module, {}).get("yaml_file", "")
        module_services = (
            ctx.modules.data.get(module, {}).get("yaml_dict", {}).get("services", {})
        )
        if not module_services:
            raise err.MinitrinoError(
                f"Invalid Docker Compose YAML file (no 'services' section found): {yaml_file}"
            )
        # Get all services defined in YAML file
        for service_key, service_dict in module_services.items():
            services.append([service_key, service_dict, yaml_file])

    containers = []
    # Get all container names for each service
    for service in services:
        bootstrap = service[1].get("environment", {}).get("MINITRINO_BOOTSTRAP")
        if bootstrap is None:
            continue
        container_name = service[1].get("container_name")
        if container_name is None:
            # If there is not container name, the service name becomes the name
            # of the container
            container_name = service[0]
        if execute_container_bootstrap(bootstrap, container_name, service[2]):
            containers.append(container_name)
    return containers


@pass_environment
def execute_container_bootstrap(ctx, bootstrap="", container_name="", yaml_file=""):
    """Executes a single bootstrap inside a container. If the
    `/opt/minitrino/bootstrap-status.txt` file has the same checksum as the
    bootstrap script that is about to be executed, the boostrap script is
    skipped.

    Returns `False` if the script is not executed and `True` if it is."""

    if any((not bootstrap, not container_name, not yaml_file)):
        raise utils.handle_missing_param(list(locals().keys()))

    bootstrap_file = os.path.join(
        os.path.dirname(yaml_file), "resources", "bootstrap", bootstrap
    )
    if not os.path.isfile(bootstrap_file):
        raise err.UserError(
            f"Bootstrap file does not exist at location: {bootstrap_file}",
            "Check this module in the library to ensure the bootstrap script is present.",
        )

    # Add executable permissions to bootstrap
    st = os.stat(bootstrap_file)
    os.chmod(
        bootstrap_file,
        st.st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH,
    )

    checksum = hashlib.md5(open(bootstrap_file, "rb").read()).hexdigest()
    container = ctx.docker_client.containers.get(container_name)

    # Check if this script has already been executed
    output = ctx.cmd_executor.execute_commands(
        "cat /opt/minitrino/bootstrap-status.txt",
        container=container,
        trigger_error=False,
    )

    if f"{checksum}" in output[0].get("output", ""):
        ctx.logger.verbose(
            f"Bootstrap already executed in container '{container_name}'. Skipping.",
        )
        return False

    ctx.logger.verbose(
        f"Executing bootstrap script in container '{container_name}'...",
    )

    ctx.cmd_executor.execute_commands(
        f"docker cp {bootstrap_file} {container_name}:/tmp/"
    )

    # Record executed file checksum
    ctx.cmd_executor.execute_commands(
        f"/tmp/{os.path.basename(bootstrap_file)}",
        f'bash -c "echo {checksum} >> /opt/minitrino/bootstrap-status.txt"',
        container=container,
    )

    ctx.logger.verbose(
        f"Successfully executed bootstrap script in container '{container_name}'.",
    )

    return True


def split_cfg(cfgs=""):
    cfgs = cfgs.strip().split("\n")
    for i, cfg in enumerate(cfgs):
        cfg = re.sub(r"\s*=\s*", "=", cfg)
        cfgs[i] = cfg.split("=", 1)
    return cfgs


@pass_environment
def get_current_trino_cfgs(ctx):
    """Get Trino config.properties and jvm.config files. Return the two sets of
    configs as lists, e.g.:

    [['a', 'b'], ['c', 'd'], ['e', 'f']]
    """

    current_cfgs = ctx.cmd_executor.execute_commands(
        f"cat {ETC_TRINO}/{TRINO_CONFIG}",
        f"cat {ETC_TRINO}/{TRINO_JVM_CONFIG}",
        container=ctx.docker_client.containers.get("trino"),
        suppress_output=True,
    )

    current_trino_cfgs = split_cfg(current_cfgs[0].get("output", ""))
    current_jvm_cfg = split_cfg(current_cfgs[1].get("output", ""))

    return current_trino_cfgs, current_jvm_cfg


@pass_environment
def write_trino_cfg(ctx, c_restart=[], modules=[]):
    """Appends Trino config from various modules if specified in the module YAML
    file. The Trino container is added to the restart list if any configs are
    written to files in the container."""

    def handle_password_authenticators(cfgs):
        merge = []
        for i, cfg in enumerate(cfgs):
            if cfg[0] == "http-server.authentication.type":
                merge.append(i)

        if not merge:
            return cfgs

        auth_property = "http-server.authentication.type="
        for i, cfg in enumerate(merge):
            if i + 1 == len(merge):
                auth_property += cfgs[cfg][1].upper()
            else:
                auth_property += f"{cfgs[cfg][1].upper()},"

        cfgs = [x for i, x in enumerate(cfgs) if i not in merge]
        cfgs.append(auth_property.split("="))
        return cfgs

    trino_container = ctx.docker_client.containers.get("trino")

    if not trino_container:
        raise err.MinitrinoError(
            f"Attempting to append Trino configuration in Trino container, "
            f"but no running Trino container was found."
        )

    cfgs = []
    jvm_cfg = []
    modules.append("trino")  # check if user placed configs in root compose yaml

    for module in modules:
        if module == "trino":
            with open(os.path.join(ctx.minitrino_lib_dir, "docker-compose.yml")) as f:
                yaml_file = yaml.load(f, Loader=yaml.FullLoader)
        else:
            yaml_file = ctx.modules.data.get(module, {}).get("yaml_dict")
        usr_cfgs = (
            yaml_file.get("services", {})
            .get("trino", {})
            .get("environment", {})
            .get("CONFIG_PROPERTIES", [])
        )
        user_jvm_cfg = (
            yaml_file.get("services", {})
            .get("trino", {})
            .get("environment", {})
            .get("JVM_CONFIG", [])
        )

        if usr_cfgs:
            cfgs.extend(split_cfg(usr_cfgs))
        if user_jvm_cfg:
            jvm_cfg.extend(split_cfg(user_jvm_cfg))

    if not cfgs and not jvm_cfg:
        return c_restart

    cfgs = handle_password_authenticators(cfgs)

    checksum_file = "/opt/minitrino/user-config.txt"
    checksum_data = bytes(str([cfgs, jvm_cfg]), "utf-8")
    checksum = hashlib.md5(checksum_data).hexdigest()

    output = ctx.cmd_executor.execute_commands(
        f"cat {checksum_file}",
        container=trino_container,
        trigger_error=False,
    )

    old_checksum = output[0].get("output", "").strip().lower()
    if not "no such file or directory" in old_checksum:
        if old_checksum == checksum:
            ctx.logger.verbose(
                "User-defined config already added to config files. Skipping...",
            )
            return c_restart

    ctx.logger.verbose(
        "Checking Trino server status before updating configs...",
    )

    retry = 0
    while retry <= 30:
        logs = trino_container.logs().decode()
        if "======== SERVER STARTED ========" in logs:
            ctx.logger.verbose(
                "Trino server started.",
            )
            break
        elif trino_container.status != "running":
            raise err.MinitrinoError(
                f"Trino container stopped running. Inspect the container logs if the "
                f"container is still available. If the container was rolled back, rerun "
                f"the command with the '--no-rollback' option, then inspect the logs."
            )
        else:
            ctx.logger.verbose(
                "Waiting for Trino server to start...",
            )
            time.sleep(1)
            retry += 1

    def append_cfgs(trino_container, usr_cfgs, current_cfgs, filename):
        """If there is an overlapping config key, replace it with the user
        config."""

        if not usr_cfgs:
            return

        current_cfgs = [
            cfg
            for cfg in current_cfgs
            if not any(cfg[0] == usr_cfg[0] for usr_cfg in usr_cfgs)
        ]

        current_cfgs.extend(usr_cfgs)
        current_cfgs = ["=".join(x) for x in current_cfgs]

        ctx.logger.verbose(
            f"Removing existing {filename} file...",
        )
        ctx.cmd_executor.execute_commands(
            f"rm {ETC_TRINO}/{filename}", container=trino_container
        )

        ctx.logger.verbose(
            f"Writing new config to {filename}...",
        )
        for current_cfg in current_cfgs:
            append_cfg = (
                f'bash -c "cat <<EOT >> {ETC_TRINO}/{filename}\n{current_cfg}\nEOT"'
            )
            ctx.cmd_executor.execute_commands(
                append_cfg, container=trino_container, suppress_output=True
            )

    ctx.logger.verbose(
        "Appending user-defined Trino config to Trino container config...",
    )

    current_trino_cfgs, current_jvm_cfg = get_current_trino_cfgs()
    append_cfgs(trino_container, cfgs, current_trino_cfgs, TRINO_CONFIG)
    append_cfgs(trino_container, jvm_cfg, current_jvm_cfg, TRINO_JVM_CONFIG)

    if not "trino" in c_restart:
        c_restart.append("trino")

    ctx.logger.verbose("Recording config checksum...")
    output = ctx.cmd_executor.execute_commands(
        f'bash -c "echo {checksum} > {checksum_file}"', container=trino_container
    )

    return c_restart


@pass_environment
def check_dup_cfgs(ctx):
    """Checks for duplicate configs in Trino config files (jvm.config and
    config.properties). This is a safety check for modules that may improperly
    modify these files."""

    def log_duplicates(cfgs, filename):
        ctx.logger.verbose(
            f"Checking Trino '{filename}' file for duplicate configs...",
        )

        unique = {}
        for cfg in cfgs:
            key = cfg[0]
            if key in unique:
                unique[key].append(cfg)
            else:
                unique[key] = [cfg]

        duplicates = ["=".join(x) for y in unique.values() for x in y if len(y) > 1]

        if duplicates:
            ctx.logger.warn(
                f"Duplicate Trino configuration properties detected in "
                f"'{filename}' file:\n{str(duplicates)}",
            )

    current_trino_cfgs, current_jvm_cfg = get_current_trino_cfgs()

    log_duplicates(current_trino_cfgs, TRINO_CONFIG)
    log_duplicates(current_jvm_cfg, TRINO_JVM_CONFIG)


@pass_environment
def restart_containers(ctx, c_restart=[]):
    """Restarts all the containers in the list."""

    if c_restart == []:
        return

    c_restart = list(set(c_restart))

    def restart_container(container_name):
        """Helper function to restart a single container."""
        try:
            container = ctx.docker_client.containers.get(container_name)
            ctx.logger.verbose(f"Restarting container '{container.name}'...")
            container.restart()
            ctx.logger.verbose(f"Container '{container.name}' restarted successfully.")
        except NotFound:
            raise err.MinitrinoError(
                f"Attempting to restart container '{container_name}', but the container was not found."
            )

    with ThreadPoolExecutor() as executor:
        futures = {
            executor.submit(restart_container, container): container
            for container in c_restart
        }

        for future in as_completed(futures):
            container_name = futures[future]
            try:
                future.result()
            except err.MinitrinoError as e:
                ctx.logger.error(
                    f"Error while restarting container '{container_name}': {str(e)}"
                )

    ctx.logger.info("All specified containers have been restarted.")


@pass_environment
def initialize_containers(ctx):
    """Initializes each container with /opt/minitrino/ directory."""

    containers = ctx.docker_client.containers.list(filters={"label": RESOURCE_LABEL})
    for container in containers:
        output = ctx.cmd_executor.execute_commands(
            "mkdir -p /opt/minitrino/",
            container=container,
            trigger_error=False,
        )
        if output[0].get("return_code", None) in [0, 126]:
            continue
        else:
            raise err.MinitrinoError(
                f"Command failed.\n"
                f"Output: {output[0].get('output', '').strip()}\n"
                f"Exit code: {output[0].get('return_code', None)}"
            )


@pass_environment
def provision_workers(ctx, c_restart=[], workers=0):
    """Provisions (or updates) workers.

    Scenarios:

    1. No workers are specified and no workers are running: do nothing.
    2. The user specifies a worker count and no workers exist: the specified
       number of workers are provisioned.
    3. The user does not specify a worker count, but there are active workers:
       the worker count will be set to the number of running workers.
    4. The user specifies a worker count greater than the number of active
       workers: the worker count will increase to the specified number.
    5. The user specifies a worker count less than the number of active workers:
       the worker with the highest iterator is removed from the network.

    I am enumerating all cases for my own sanity.
    """

    # Check for running worker containers
    containers = ctx.docker_client.containers.list()
    worker_containers = [
        c.name
        for c in containers
        if c.name.startswith("trino-worker-")
        and c.labels.get("com.starburst.tests") == "minitrino"
    ]

    # Scenario 1
    if workers == 0 and len(worker_containers) == 0:
        return c_restart
    # Scenario 2
    if workers > len(worker_containers):
        pass
    # Scenario 3
    if workers == 0 and len(worker_containers) > 0:
        workers = len(worker_containers)
    # Scenario 4
    if workers > len(worker_containers):
        pass
    # Scenario 5
    if workers < len(worker_containers):
        worker_containers.sort(reverse=True)
        excess = len(worker_containers) - workers
        remove = worker_containers[:excess]
        for c in remove:
            c = ctx.docker_client.containers.get(c)
            c.kill()
            c.remove()
            identifier = utils.generate_identifier({"ID": c.short_id, "Name": c.name})
            ctx.logger.warn(f"Removed excess worker: {identifier}")

    worker_img = f"minitrino/trino:{ctx.env.get('STARBURST_VER')}"
    source_directory = "/etc/starburst"
    network_name = "minitrino_default"

    coordinator = ctx.docker_client.containers.get("trino")

    for i in range(1, workers + 1):
        worker_name = f"trino-worker-{i}"
        try:
            worker = ctx.docker_client.containers.get(worker_name)
        except NotFound:
            worker = ctx.docker_client.containers.run(
                worker_img,
                name=worker_name,
                detach=True,
                network=network_name,
                labels={
                    "com.starburst.tests": "minitrino",
                    "com.starburst.tests.module": "trino",
                },
            )
            ctx.logger.verbose(
                f"Created and started worker container: '{worker_name}' in network '{network_name}'"
            )

        # Copy the source directory from the coordinator to the worker container
        tar_path = "/tmp/starburst.tar.gz"
        ctx.cmd_executor.execute_commands(
            f"tar czf {tar_path} -C /etc starburst",
            container=coordinator,
            docker_user="starburst",
        )

        # Copy the tar file from the coordinator container
        bits, _ = coordinator.get_archive(tar_path)
        tar_stream = b"".join(bits)

        # Put the tar file into the new worker container & extract
        worker.put_archive("/tmp", tar_stream)
        ctx.cmd_executor.execute_commands(
            f"tar xzf /tmp/starburst.tar.gz -C /etc",
            container=worker,
            docker_user="starburst",
        )

        # Overwrite worker config.properties
        ctx.cmd_executor.execute_commands(
            f"bash -c \"echo '{WORKER_CONFIG_PROPS}' > {ETC_TRINO}/config.properties\"",
            container=worker,
            docker_user="starburst",
        )

        c_restart.append(worker_name)
        ctx.logger.verbose(f"Copied {source_directory} to '{worker_name}'")

    return c_restart


@pass_environment
def rollback(ctx, no_rollback):
    """Rolls back the provisioning command in the event of an error."""

    if no_rollback:
        ctx.logger.warn(
            f"Errors occurred during environment provisioning and rollback has been disabled. "
            f"Provisioned resources will remain in an unaltered state.",
        )
        return

    ctx.logger.warn(
        f"Rolling back provisioned resources due to "
        f"errors encountered while provisioning the environment.",
    )

    containers = ctx.docker_client.containers.list(
        filters={"label": RESOURCE_LABEL}, all=True
    )

    for container in containers:
        container.kill()
        container.remove()



#!usr/bin/env/python3
# -*- coding: utf-8 -*-

import sys
import click

from minitrino.cli import pass_environment
from minitrino import utils
from minitrino.settings import IMAGE
from minitrino.settings import VOLUME
from minitrino.settings import RESOURCE_LABEL
from docker.errors import APIError


@click.command(
    "remove",
    help=("""Remove Minitrino resources."""),
)
@click.option(
    "-i",
    "--images",
    is_flag=True,
    default=False,
    help=("""Remove Minitrino images."""),
)
@click.option(
    "-v",
    "--volumes",
    is_flag=True,
    default=False,
    help=("""Remove Minitrino container volumes."""),
)
@click.option(
    "-l",
    "--label",
    "labels",
    type=str,
    default=[],
    multiple=True,
    help=(
        """Target specific labels for removal (format: key-value
        pair(s))."""
    ),
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help=(
        """Force the removal of Minitrino resources. Normal Docker removal
        restrictions apply."""
    ),
)
@utils.exception_handler
@pass_environment
def cli(ctx, images, volumes, labels, force):
    """Remove command for Minitrino."""

    utils.check_daemon(ctx.docker_client)

    if all((not images, not volumes, not labels)) or all((images, volumes, not labels)):
        response = ctx.logger.prompt_msg(
            "You are about to all remove minitrino images and volumes. Continue? [Y/N]"
        )
        if utils.validate_yes(response):
            remove_items(IMAGE, force)
            remove_items(VOLUME, force)
        else:
            ctx.logger.info(f"Opted to skip resource removal.")
            sys.exit(0)

    if images:
        remove_items(IMAGE, force, labels)
    if volumes:
        remove_items(VOLUME, force, labels)

    ctx.logger.info(f"Removal complete.")


@pass_environment
def remove_items(ctx, item_type, force, labels=[]):
    """Removes Docker items. If no labels are passed in, all Minitrino
    resources are removed. If label(s) are passed in, the removal is limited to
    the passed in labels."""

    if not labels:
        labels = [RESOURCE_LABEL]

    images = []
    volumes = []
    for label in labels:
        if item_type == IMAGE:
            images.extend(ctx.docker_client.images.list(filters={"label": label}))
        if item_type == VOLUME:
            volumes.extend(ctx.docker_client.volumes.list(filters={"label": label}))

    images = list(set(images))
    for image in images:
        try:
            identifier = utils.generate_identifier(
                {"ID": image.short_id, "Image:Tag": try_get_image_tag(image)}
            )
            if force:
                ctx.docker_client.images.remove(
                    image.short_id, force=True, noprune=False
                )
            else:
                ctx.docker_client.images.remove(image.short_id)
            ctx.logger.verbose(
                f"{item_type.title()} removed: {identifier}",
            )
        except APIError as e:
            ctx.logger.verbose(
                f"Cannot remove image: {identifier}\n"
                f"Error from Docker: {e.explanation}",
            )

    volumes = list(set(volumes))
    for volume in volumes:
        try:
            identifier = utils.generate_identifier({"ID": volume.id})
            if force:
                volume.remove(force=True)
            else:
                volume.remove()
            ctx.logger.verbose(
                f"{item_type.title()} removed: {identifier}",
            )
        except APIError as e:
            ctx.logger.verbose(
                f"Cannot remove volume: {identifier}\n"
                f"Error from Docker: {e.explanation}",
            )


def try_get_image_tag(image):
    """Tries to get an image tag. If there is no tag, returns an empty string."""

    try:
        return image.tags[0]
    except:
        return ""



#!usr/bin/env/python3
# -*- coding: utf-8 -*-

import os
import json
import re
import yaml
import docker
import subprocess

from pathlib import Path
from configparser import ConfigParser

from minitrino import utils
from minitrino import errors as err
from minitrino.settings import RESOURCE_LABEL
from minitrino.settings import MODULE_ROOT
from minitrino.settings import MODULE_ADMIN
from minitrino.settings import MODULE_SECURITY
from minitrino.settings import MODULE_CATALOG


class Environment:
    """Provides context and core controls that are globally accessible in
    command scripts. This class should not be instantiated from anywhere but the
    CLI's entrypoint, as it depends on user-provided inputs.

    ### Public Attributes (Interactive)
    - `logger`: A `minitrino.utils.Logger` object.
    - `env`: An `EnvironmentVariables` object containing all CLI environment
        variables, subdivided by sections when possible.
    - `modules`: A `Modules` object containing metadata about Minitrino
      modules.
    - `cmd_executor`: A `CommandExecutor` object to execute shell commands in
      the host shell and inside containers.
    - `docker_client`: A `docker.DockerClient` object.
    - `api_client`: A `docker.APIClient` object.

    ### Public Attributes (Static)
    - `verbose`: If `True`, logs flagged as verbose to are sent to stdout.
    - `user_home_dir`: The home directory of the current user.
    - `minitrino_user_dir`: The location of the Minitrino directory relative
      to the user home directory (~/.minitrino/).
    - `config_file`: The location of the user's minitrino.cfg file.
    - `snapshot_dir`: The location of the user's snapshot directory (this is
        essentially a temporary directory, as 'permanent' snapshot tarballs are
        written to the library or user-specified directory).
    - `minitrino_lib_dir`: The location of the Minitrino library."""

    @utils.exception_handler
    def __init__(self):
        # Attributes that depend on user input prior to being set
        self.verbose = False
        self._user_env = []

        self.logger = utils.Logger()
        self.env = EnvironmentVariables
        self.modules = Modules
        self.cmd_executor = CommandExecutor
        self.docker_client = docker.DockerClient
        self.api_client = docker.APIClient

        # Paths
        self.user_home_dir = os.path.expanduser("~")
        self.minitrino_user_dir = self._handle_minitrino_user_dir()
        self.config_file = self._get_config_file()
        self.snapshot_dir = os.path.join(self.minitrino_user_dir, "snapshots")

    @property
    def minitrino_lib_dir(self):
        """The directory of the Minitrino library. The directory can be
        determined in four ways (this is the order of precedence):
        1. Pass `LIB_PATH` to the CLI via `-e`/ `--env`.
        2. Export `LIB_PATH` as a shell environment variable.
        3. Set `LIB_PATH` in `minitrino.cfg`.
        4. As a last resort, Minitrino will check to see if the library exists
           in relation to the positioning of the `components.py` file and
           assumes the project is being run out of a cloned repository."""

        lib_dir = ""
        try:
            lib_dir = self.env.get("LIB_PATH", "")
        except:
            pass

        if not lib_dir and os.path.isdir(os.path.join(self.minitrino_user_dir, "lib")):
            lib_dir = os.path.join(self.minitrino_user_dir, "lib")
        elif not lib_dir:  # Use repo root, fail if this doesn't exist
            lib_dir = Path(os.path.abspath(__file__)).resolve().parents[2]
            lib_dir = os.path.join(lib_dir, "lib")

        if not os.path.isdir(lib_dir) or not os.path.isfile(
            os.path.join(lib_dir, "minitrino.env")
        ):
            raise err.UserError(
                "You must provide a path to a compatible Minitrino library ",
                f"via `LIB_PATH` or by placing the library in the default location, "
                f"~/.minitrino/lib",
            )
        return lib_dir

    @utils.exception_handler
    def _user_init(self, verbose=False, user_env=[]):
        """Initialize attributes that depend on user-provided input."""

        # Update static attributes
        self.verbose = verbose
        self._user_env = user_env

        # Instantiate/update interactive attributes
        self.logger = utils.Logger(self.verbose)
        self.env = EnvironmentVariables(self)

        # Skip the library-related procedures if the library is not found
        try:
            if self.minitrino_lib_dir:
                self.logger.verbose(
                    f"Library path set to: {self.minitrino_lib_dir}",
                )

            # Now that we know where the library is, we can try to parse the env
            # file and obtain all of the modules
            self.env._parse_library_env()
            self.modules = Modules(self)

            # Warn the user if the library and CLI vers don't match
            cli_ver = utils.get_cli_ver()
            lib_ver = utils.get_lib_ver(self.minitrino_lib_dir)
            if cli_ver != lib_ver:
                self.logger.warn(
                    f"CLI version {cli_ver} and library version {lib_ver} "
                    f"do not match. You can update the Minitrino library "
                    f"version to match the CLI version by running 'minitrino "
                    f"lib-install'.",
                )
        except:
            pass

        self.env._log_env_vars()
        self.cmd_executor = CommandExecutor(self)
        self._get_docker_clients()

    def _handle_minitrino_user_dir(self):
        """Checks if a Minitrino directory exists in the user home directory.
        If it does not, it is created. The path to the Minitrino user home
        directory is returned."""

        minitrino_user_dir = os.path.abspath(
            os.path.join(self.user_home_dir, ".minitrino")
        )
        if not os.path.isdir(minitrino_user_dir):
            os.mkdir(minitrino_user_dir)
        return minitrino_user_dir

    def _get_config_file(self):
        """Returns the correct filepath for the minitrino.cfg file. Adds to
        initialization warnings if the file does not exist, but will return the
        path regardless."""

        config_file = os.path.join(self.minitrino_user_dir, "minitrino.cfg")
        if not os.path.isfile(config_file):
            self.logger.warn(
                f"No minitrino.cfg file found at {config_file}. "
                f"Run 'minitrino config' to reconfigure this file and directory.",
            )
        return config_file

    def _get_docker_clients(self):
        """Gets DockerClient and APIClient objects. Returns a tuple of DockerClient
        and APIClient objects, respectively.

        If there is an error fetching the clients, None types will be returned
        for each client. The lack of clients should be caught by check_daemon()
        calls that execute in each command that requires an accessible Docker
        service."""

        docker_url = os.environ.get("DOCKER_HOST", "")

        try:
            docker_client = docker.DockerClient(base_url=docker_url)
            api_client = docker.APIClient(base_url=docker_url)
            self.docker_client, self.api_client = docker_client, api_client
        except:
            return None, None


class EnvironmentVariables(dict):
    """Exposes all Minitrino variables.

    ### Parameters
    - `ctx`: Instantiated Environment object (with user input already accounted
      for).

    ### Usage
    ```python
    # ctx object has an instantiated EnvironmentVariables object
    env_variable = ctx.env.get("STARBURST_VER", "388-e")
    ```"""

    @utils.exception_handler
    def __init__(self, ctx=None):
        super().__init__()
        if not ctx:
            raise utils.handle_missing_param(list(locals().keys()))

        self._ctx = ctx

        self._parse_user_env()
        self._parse_os_env()
        self._parse_minitrino_config()

    def _parse_user_env(self):
        """Parses user-provided environment variables for the current
        command. Highest precedence in environment variable order, loaded
        first."""

        if not self._ctx._user_env:
            return

        for env_var in self._ctx._user_env:
            env_var = utils.parse_key_value_pair(env_var, err_type=err.UserError)
            self[env_var[0]] = env_var[1]

    def _parse_os_env(self):
        """Parses environment variables from the user's shell. Middle
        precedence in environment variable order, loaded second."""

        append = ["LIB_PATH", "STARBURST_VER", "TEXT_EDITOR", "LIC_PATH"]
        for k, v in os.environ.items():
            k = k.upper()
            if k in append and not self.get(k, None):
                self[k] = v

    def _parse_minitrino_config(self):
        """Parses the Minitrino config file and adds it to the env
        dictionary. Middle precedence in environment variable order, loaded
        third."""

        if not os.path.isfile(self._ctx.config_file):
            return

        try:
            config = ConfigParser()
            config.optionxform = str  # Preserve case
            config.read(self._ctx.config_file)
            for k, v in config.items("config"):
                if not self.get(k, None) and v:
                    self[k.upper()] = v
        except Exception as e:
            utils.handle_exception(
                e,
                additional_msg=f"Failed to parse config file: {self._ctx.config_file}",
            )

    def _parse_library_env(self):
        """Parses the Minitrino library's `minitrino.env` file. Lowest
        precedence in environment variable order, loaded last."""

        env_file = os.path.join(self._ctx.minitrino_lib_dir, "minitrino.env")
        if not os.path.isfile(env_file):
            raise err.UserError(
                f"Library 'minitrino.env' file does not exist at path: {env_file}",
                f"Are you pointing to a valid library, and is the minitrino.env file "
                f"present in that library?",
            )

        with open(env_file, "r") as f:
            for env_var in f:
                env_var = utils.parse_key_value_pair(env_var, err_type=err.UserError)
                if env_var is None:
                    continue
                if not self.get(env_var[0], None):
                    self[env_var[0]] = env_var[1]

    def _log_env_vars(self):
        """Logs environment variables."""

        if self:
            self._ctx.logger.verbose(
                f"Registered environment variables:\n{json.dumps(self, indent=2)}",
            )


class Modules:
    """Contains information about all valid Minitrino modules.

    ### Parameters
    - `ctx`: Instantiated Environment object (with user input already accounted
      for).

    ### Public Attributes
    - `data`: A dictionary of module information.

    ### Public Methods
    - `get_running_modules()`: Returns a list of running modules."""

    @utils.exception_handler
    def __init__(self, ctx=None):
        if not ctx:
            raise utils.handle_missing_param(list(locals().keys()))

        self.data = {}
        self._ctx = ctx
        self._load_modules()

    def get_running_modules(self):
        """Returns dict of running modules (includes container objects and
        Docker labels)."""

        utils.check_daemon(self._ctx.docker_client)
        containers = self._ctx.docker_client.containers.list(
            filters={"label": RESOURCE_LABEL}
        )

        if not containers:
            return []

        modules = []
        for container in containers:
            label_set = {}
            ids = ["admin-", "catalog-", "security-"]
            for k, v in container.labels.items():
                for _id in ids:
                    if "com.starburst.tests" in k and _id in v:
                        modules.append(v.lower().strip().replace(_id, ""))
                        label_set[k] = v
            # All containers except the trino containers must have
            # module-specific labels. The trino container only has module labels
            # if a module applies labels to it
            if not label_set and not (
                "trino-worker" in container.name or container.name == "trino"
            ):
                raise err.UserError(
                    f"Missing Minitrino labels for container '{container.name}'.",
                    f"Check this module's 'docker-compose.yml' file and ensure you are "
                    f"following the documentation on labels.",
                )

        modules = list(set(modules))
        for module in modules:
            if not isinstance(self.data.get(module), dict):
                raise err.UserError(
                    f"Module '{module}' is running, but it is not found "
                    f"in the library. Was it deleted, or are you pointing "
                    f"Minitrino to the wrong location?"
                )
        return modules

    def _load_modules(self):
        """Loads module data during instantiation."""

        self._ctx.logger.verbose("Loading modules...")

        modules_dir = os.path.join(self._ctx.minitrino_lib_dir, MODULE_ROOT)
        if not os.path.isdir(modules_dir):
            raise err.MinitrinoError(
                f"Path is not a directory: {modules_dir}. "
                f"Are you pointing to a compatible Minitrino library?"
            )

        # Loop through all module types
        sections = [
            os.path.join(modules_dir, MODULE_ADMIN),
            os.path.join(modules_dir, MODULE_CATALOG),
            os.path.join(modules_dir, MODULE_SECURITY),
        ]

        for section_dir in sections:
            for _dir in os.listdir(section_dir):
                module_dir = os.path.join(section_dir, _dir)

                if not os.path.isdir(module_dir):
                    self._ctx.logger.verbose(
                        f"Skipping file (expected a directory, not a file) "
                        f"at path: {module_dir}",
                    )
                    continue

                # List inner-module files
                module_files = os.listdir(module_dir)

                yaml_basename = f"{os.path.basename(module_dir)}.yml"
                if not yaml_basename in module_files:
                    raise err.UserError(
                        f"Missing Docker Compose file in module directory {_dir}. "
                        f"Expected file to be present: {yaml_basename}",
                        hint_msg="Check this module in your library to ensure it is properly constructed.",
                    )

                # Module dir and YAML exist, add to modules
                module_name = os.path.basename(module_dir)
                self.data[module_name] = {}
                self.data[module_name]["type"] = os.path.basename(section_dir)
                self.data[module_name]["module_dir"] = module_dir

                # Add YAML file path
                yaml_file = os.path.join(module_dir, yaml_basename)
                self.data[module_name]["yaml_file"] = yaml_file

                # Add YAML dict
                with open(yaml_file) as f:
                    self.data[module_name]["yaml_dict"] = yaml.load(
                        f, Loader=yaml.FullLoader
                    )

                # Get metadata.json if present
                json_basename = "metadata.json"
                json_file = os.path.join(module_dir, json_basename)
                metadata = {}
                if os.path.isfile(json_file):
                    with open(json_file) as f:
                        metadata = json.load(f)
                else:
                    self._ctx.logger.verbose(
                        f"No JSON metadata file for module '{module_name}'. "
                        f"Will not load metadata for module.",
                    )
                for k, v in metadata.items():
                    self.data[module_name][k] = v


class CommandExecutor:
    """Executes commands in the host shell/host containers with customized
    handling of stdout/stderr output.

    ### Parameters
    - `ctx`: Instantiated Environment object (with user input already accounted
      for).

    ### Public Methods
    - `execute_commands()`: Executes commands in the user's shell or inside of a
        container."""

    @utils.exception_handler
    def __init__(self, ctx=None):
        if not ctx:
            raise utils.handle_missing_param(list(locals().keys()))

        self._ctx = ctx

    def execute_commands(self, *args, **kwargs):
        """Executes commands in the user's shell or inside of a container.
        Returns output as well as stores the output in the `output` attribute.

        ### Parameters
        - `args`: Commands that will be executed in the order provided.

        Keyword Arguments:

        - `trigger_error`: If `False`, errors (non-zero exit codes) from
          executed commands will not raise an exception. Defaults to `True`.
        - `environment`: A dictionary of environment variables to pass to the
          subprocess or container.
        - `suppress_output`: If `True`, output to stdout from the executed
          command will be suppressed.
        - `container`: A Docker container object. If passed in, the command will
          be executed through the Docker SDK instead of the subprocess module.
        - `docker_user`: The user to execute the command as in the Docker
          container (default: `root`).

        ### Return Values
        - A list of dicts with each dict containing the following keys:
            - `command`: the original command passed to the function
            - `output`: the combined output of stdout and stderr
            - `return_code`: the return code of the command"""

        output = []
        if kwargs.get("container", None):
            kwargs["environment"] = self._construct_environment(
                kwargs.get("environment", {}), kwargs.get("container", None)
            )
            for command in args:
                output.append(self._execute_in_container(command, **kwargs))
        else:
            kwargs["environment"] = self._construct_environment(
                kwargs.get("environment", {})
            )
            for command in args:
                output.append(self._execute_in_shell(command, **kwargs))

        return output

    def _execute_in_shell(self, command="", **kwargs):
        """Executes a command in the host shell."""

        self._ctx.logger.verbose(
            f"Executing command in shell:\n{command}",
        )

        process = subprocess.Popen(
            command,
            shell=True,
            env=kwargs.get("environment", {}),
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )

        if not kwargs.get("suppress_output", False):
            # Stream the output of the executed command line-by-line.
            # `universal_newlines=True` ensures output is generated as a string,
            # so there is no need to decode bytes. The only cleansing we need to
            # do is to run the string through the `_strip_ansi()` function.

            started_stream = False
            output = ""
            while True:
                output_line = process.stdout.readline()
                if output_line == "" and process.poll() is not None:
                    break
                output_line = self._strip_ansi(output_line)
                if not started_stream:
                    self._ctx.logger.verbose("Command Output:")
                    started_stream = True
                self._ctx.logger.verbose(output_line, stream=True)
                output += output_line

        if process.returncode != 0 and kwargs.get("trigger_error", True):
            raise err.MinitrinoError(
                f"Failed to execute shell command:\n{command}\n"
                f"Exit code: {process.returncode}"
            )

        return {
            "command": command,
            "output": self._strip_ansi(output),
            "return_code": process.returncode,
        }

    def _execute_in_container(self, command="", **kwargs):
        """Executes a command inside of a container through the Docker SDK
        (similar to `docker exec`)."""

        container = kwargs.get("container", None)
        if container is None:
            raise err.MinitrinoError(
                f"Attempted to execute a command inside of a "
                f"container, but a container object was not provided."
            )

        if not kwargs.get("suppress_output"):
            self._ctx.logger.verbose(
                f"Executing command in container '{container.name}':\n{command}",
            )

        # Create exec handler and execute the command
        exec_handler = self._ctx.api_client.exec_create(
            container.name,
            cmd=command,
            environment=kwargs.get("environment", {}),
            privileged=True,
            user=kwargs.get("docker_user", "root"),
        )

        # `output` is a generator that yields response chunks
        output_generator = self._ctx.api_client.exec_start(exec_handler, stream=True)

        # Output from the generator is returned as bytes, so they need to be
        # decoded to strings. Response chunks are not guaranteed to be full
        # lines. A newline in the output chunk will trigger a log dump of the
        # current `full_line` up to the first newline in the current chunk. The
        # remainder of the chunk (if any) resets the `full_line` var, then log
        # dumped when the next newline is received.

        output = ""
        full_line = ""
        started_stream = False
        for chunk in output_generator:
            chunk = self._strip_ansi(chunk.decode())
            output += chunk
            chunk = chunk.split("\n", 1)
            if len(chunk) > 1:  # Indicates newline present
                full_line += chunk[0]
                if not kwargs.get("suppress_output", False):
                    if not started_stream:
                        self._ctx.logger.verbose("Command Output:")
                        started_stream = True
                    self._ctx.logger.verbose(full_line, stream=True)
                    full_line = ""
                if chunk[1]:
                    full_line = chunk[1]
            else:
                full_line += chunk[0]

        # Catch lingering full line post-loop
        if not kwargs.get("suppress_output", False) and full_line:
            self._ctx.logger.verbose(full_line, stream=True)

        # Get the exit code
        return_code = self._ctx.api_client.exec_inspect(exec_handler["Id"]).get(
            "ExitCode"
        )
        # https://www.gnu.org/software/bash/manual/html_node/Exit-Status.html
        if return_code == 126:
            self._ctx.logger.warn(
                f"The command exited with a 126 code which typically means an "
                f"executable is not accessible or installed. Does this image have "
                f"all required dependencies installed?\nCommand: {command}",
            )

        if return_code != 0 and kwargs.get("trigger_error", True):
            raise err.MinitrinoError(
                f"Failed to execute command in container '{container.name}':\n{command}\n"
                f"Exit code: {return_code}"
            )

        return {"command": command, "output": output, "return_code": return_code}

    def _construct_environment(self, environment={}, container=None):
        """Merges provided environment dictionary with user's shell environment
        variables. For shell execution, the host environment will be set to the
        existing variables in the host environment. For container execution, the
        host environment will be set to the container's existing environment
        variables."""

        # Remove conflicting keys from host environment; Minitrino environment
        # variables take precendance

        if not container:
            host_environment = os.environ.copy()
        else:
            host_environment_list = self._ctx.api_client.inspect_container(
                container.id
            )["Config"]["Env"]
            host_environment = {}
            for env_var in host_environment_list:
                env_var = utils.parse_key_value_pair(env_var)
                host_environment[env_var[0]] = env_var[1]

        if environment:
            delete_keys = []
            for host_key, host_value in host_environment.items():
                for key, value in environment.items():
                    if key == host_key:
                        delete_keys.append(host_key)
            for delete_key in delete_keys:
                del host_environment[delete_key]

        # Merge environment argument with copy of existing environment
        environment.update(host_environment)
        return environment

    def _strip_ansi(self, value=""):
        """Strips ANSI escape sequences from strings."""

        # Strip ANSI codes before Click so that our logging helpers
        # know if it's an empty string or not.
        ansi_regex = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
        return ansi_regex.sub("", value)



#!usr/bin/env/python3
# -*- coding: utf-8 -*-

# Docker labels
RESOURCE_LABEL = "com.starburst.tests=minitrino"
MODULE_LABEL_KEY_ROOT = "com.starburst.tests.module"

# Generic Constants
CONTAINER = "container"
IMAGE = "image"
VOLUME = "volume"
LIB = "lib"
MODULE_ROOT = "modules"
MODULE_ADMIN = "admin"
MODULE_CATALOG = "catalog"
MODULE_SECURITY = "security"
MODULE_RESOURCES = "resources"
MIN_SEP_VER = 402
ETC_TRINO = "/etc/starburst"
LIC_VOLUME_MOUNT = "${LIC_PATH}:${LIC_MOUNT_PATH}"
LIC_MOUNT_PATH = "/etc/starburst/starburstdata.license:ro"
DUMMY_LIC_MOUNT_PATH = "/etc/starburst/dummy.license:ro"
TRINO_CONFIG = "config.properties"
TRINO_JVM_CONFIG = "jvm.config"
LIB_INDEPENDENT_CMDS = ["lib-install"]

# Snapshots
SNAPSHOT_ROOT_FILES = [
    "docker-compose.yml",
    "minitrino.env",
    "version",
    "image",
]

# Terminal
DEFAULT_INDENT = " " * 5

# Scrub Keys
SCRUB_KEYS = [
    "accesskey",
    "apikey",
    "secretkey",
    "-key",
    "_key",
    "password",
]

# Templates
CONFIG_TEMPLATE = """
[config]
# defaults to ~/.minitrino/lib
LIB_PATH=

STARBURST_VER=
TEXT_EDITOR=
LIC_PATH=
"""

WORKER_CONFIG_PROPS = """coordinator=false
http-server.http.port=8080
discovery.uri=http://trino:8080
internal-communication.shared-secret=bWluaXRyaW5vUm9ja3MxNQo="""

PROVISION_SNAPSHOT_TEMPLATE = """
#!/usr/bin/env bash

# ------------------------------------------------------------------------------------
# Below is the exact command used to provision the snapshotted environment. Run this 
# command in your terminal to reproduce the exact state of the environment.
#
# If you need config data from the snapshot's 'minitrino.cfg' file, you will either 
# need to copy it from the snapshot directory to '~./minitrino/minitrino.cfg' or 
# individually copy the needed configs to your existing 'minitrino.cfg' file. 
# ------------------------------------------------------------------------------------


"""



#!usr/bin/env/python3
# -*- coding: utf-8 -*-

from minitrino import utils


class MinitrinoError(Exception):
    """Generic Minitrino exception class."""

    exit_code = 1

    def __init__(self, msg=""):
        if not msg:
            raise utils.handle_missing_param(list(locals().keys()))
        super().__init__(msg)
        self.msg = msg

    def __str__(self):
        return self.msg


class UserError(MinitrinoError):
    """An exception that Minitrino can handle and show to the user.

    ### Parameters
    - `msg`: The message to log, and the raised exception's message.
    - `hint_msg`: User hint (optional). Should help the user figure out how to
      solve the problem."""

    exit_code = 2

    def __init__(self, msg="", hint_msg=""):
        if not msg:
            raise utils.handle_missing_param("msg")
        if hint_msg:
            super().__init__(f"User error: {msg}\nHint: {hint_msg}")
        else:
            super().__init__(f"User error: {msg}")



#!usr/bin/env/python3
# -*- coding: utf-8 -*-

import os
import sys
import traceback
import pkg_resources

from minitrino import errors as err
from minitrino.settings import DEFAULT_INDENT
from minitrino.settings import MIN_SEP_VER

from click import echo, style, prompt
from textwrap import fill
from shutil import get_terminal_size
from functools import wraps


class Logger:
    """Minitrino logging class. The logger does not log to a file; it uses
    `click.echo()` to print text to the user's terminal.

    The log level will affect the prefix color (i.e. the '[i]' in info messages
    is blue) and the message prefix (the prefix for the warning logs is '[w]').

    ### Parameters
    - `log_verbose`: If `True`, log messages flagged as verbose will be logged.
      If `False`, verbose messages will not be logged. This is dynamically
      determined by user input when the logger is instantiated from the
      Environment class.

    ### Attributes
    - `INFO`: Info log level.
    - `WARN`: Warn log level.
    - `ERROR`: Error log level.
    - `VERBOSE`: Verbose log level.

    ### Methods
    - `log()`: Logs a message to the user's terminal. Level must be specified,
      defaults to INFO.
    - `info()`: Logs an info message
    - `warn()`: Logs a warning message
    - `error()`: Logs an error message
    - `verbose()`: Logs a verbose message
    - `prompt_msg()`: Logs a prompt message and returns the user's input."""

    def __init__(self, log_verbose=False):
        self.INFO = {"prefix": "[i]  ", "prefix_color": "cyan"}
        self.WARN = {"prefix": "[w]  ", "prefix_color": "yellow"}
        self.ERROR = {"prefix": "[e]  ", "prefix_color": "red"}
        self.VERBOSE = {"prefix": "[i]  ", "prefix_color": "cyan", "verbose": True}

        self._log_verbose = log_verbose

    def log(self, *args, level=None, stream=False):
        """Logs messages to the user's console. Defaults to 'info' log level.

        ### Parameters
        - `*args`: Messages to log.
        - `level`: The level of the log message (info, warn, error, and
          verbose).
        - `stream`: If `True`, the logger will not apply a prefix to each line
          streamed to the console."""

        if not level:
            level = self.INFO

        # Skip verbose messages unless verbose mode is enabled
        if not self._log_verbose and level == self.VERBOSE:
            return

        for msg in args:
            # Ensure the message can be a string
            try:
                msg = str(msg)
            except:
                raise err.MinitrinoError(
                    f"A string is required for {self.log.__name__}."
                )
            msgs = msg.replace("\r", "\n").split("\n")
            # Log each message
            for i, msg in enumerate(msgs):
                msg = self._format(msg)
                if not msg:
                    continue
                if stream or i > 0:
                    msg_prefix = DEFAULT_INDENT
                else:
                    msg_prefix = style(
                        level.get("prefix", ""),
                        fg=level.get("prefix_color", ""),
                        bold=True,
                    )
                echo(f"{msg_prefix}{msg}")

    def info(self, *args, stream=False):
        self.log(*args, level=self.INFO, stream=stream)

    def warn(self, *args, stream=False):
        self.log(*args, level=self.WARN, stream=stream)

    def error(self, *args, stream=False):
        self.log(*args, level=self.ERROR, stream=stream)

    def verbose(self, *args, stream=False):
        self.log(*args, level=self.VERBOSE, stream=stream)

    def prompt_msg(self, msg="", input_type=str):
        """Logs a prompt message and returns the user's input.

        ### Parameters
        - `msg`: The prompt message
        - `input_type`: The object type to check the input for"""

        if not msg:
            raise handle_missing_param(["msg"])

        try:
            msg = str(msg)
        except:
            raise err.MinitrinoError(f"A string is required for {self.log.__name__}.")

        msg = self._format(msg)
        styled_prefix = style(
            self.INFO.get("prefix", ""), fg=self.INFO.get("prefix_color", ""), bold=True
        )

        return prompt(
            f"{styled_prefix}{msg}",
            type=input_type,
        )

    def _format(self, msg):
        """Formats strings prior to displaying to the user."""

        msg = msg.rstrip()
        if not msg:
            return ""

        terminal_width, _ = get_terminal_size()
        msg = msg.replace("\n", f"\n{DEFAULT_INDENT}")
        msg = fill(
            msg,
            terminal_width - 4,
            subsequent_indent=DEFAULT_INDENT,
            replace_whitespace=False,
            break_on_hyphens=False,
            break_long_words=True,
        )

        return msg


def handle_exception(error=Exception, additional_msg="", skip_traceback=False):
    """Handles a single exception. Wrapped by `@exception_handler` decorator.

    ### Parameters
    - `error`: The exception object.
    - `additional_msg`: An additional message to log, if any. Can be useful if
      handling a generic exception and you need to append a user-friendly
      message to the log.
    - `skip_traceback`: If `True`, the traceback will not be printed to the
      user's terminal. Defaults to `True` for user errors, but it is `False`
      otherwise."""

    if not isinstance(error, Exception):
        raise handle_missing_param(["error"])

    if isinstance(error, err.UserError):
        error_msg = error.msg
        exit_code = error.exit_code
        skip_traceback = True
    elif isinstance(error, err.MinitrinoError):
        error_msg = error.msg
        exit_code = error.exit_code
    elif isinstance(error, Exception):
        error_msg = str(error)
        exit_code = 1
    else:
        raise err.MinitrinoError(
            f"Invalid type given to 'e' parameter of {handle_exception.__name__}. "
            f"Expected an Exception type, but got type {type(error).__name__}"
        )

    logger = Logger()
    logger.error(additional_msg, error_msg)
    if not skip_traceback:
        echo()  # Force a newline
        echo(f"{traceback.format_exc()}", err=True)

    sys.exit(exit_code)


def exception_handler(func):
    """A decorator that handles unhandled exceptions. Why? A few reasons.

    1. Inner functions still have the liberty to do try/catch and perform
       inner-function exception handling how they wish.
    2. Functions that catch exceptions which need specialized handling can
       manually invoke the handle_exception() utility.
    3. For all other generic exceptions and unhandled exceptions, they will be
       siphoned to the handle_exception() utility.

    This is especially useful when decorating main/runner functions and class
    constructors."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            handle_exception(e)

    return wrapper


def handle_missing_param(params=[]):
    """Handles missing parameters required for function calls. This should be
    used to signal a programmatic error, not a user error.

    ### Parameters
    - `params`: List of parameter names that are required.

    ### Usage
    ```python
    # All params are required
    if not param:
        raise handle_missing_param(list(locals().keys()))
    # Two params are required
    if not param:
        raise handle_missing_param(["module", "path"])
    ```"""

    if not params:
        raise handle_missing_param(list(locals().keys()))

    return err.MinitrinoError(f"Parameters {params} required to execute function.")


def check_daemon(docker_client):
    """Checks if the Docker daemon is running. If an exception is thrown, it is
    handled."""

    try:
        docker_client.ping()
    except Exception as e:
        raise err.UserError(
            f"Error when pinging the Docker server. Is the Docker daemon running?\n"
            f"Error from Docker: {str(e)}",
            f"You may need to initialize your Docker daemon. If Docker is already "
            f"running, check if you are using a non-default Docker context (e.g. Colima), "
            f"and if so, add `DOCKER_HOST` to your `bash_profile` and point to the relevant "
            f"`.sock` file. For more help, visit: \n"
            f"https://github.com/jefflester/minitrino/wiki/Installation-and-Upgrades",
        )


def check_lib(ctx):
    """Checks if a Minitrino library exists."""

    ctx.minitrino_lib_dir


def check_starburst_ver(ctx):
    """Checks if a proper Starburst version is provided."""

    starburst_ver = ctx.env.get("STARBURST_VER", "")
    error_msg = (
        f"Provided Starburst version '{starburst_ver}' is invalid. "
        f"The provided version must be {MIN_SEP_VER}-e or higher."
    )

    try:
        starburst_ver_int = int(starburst_ver[0:3])
        if starburst_ver_int < MIN_SEP_VER or "-e" not in starburst_ver:
            raise err.UserError(error_msg)
    except:
        raise err.UserError(error_msg)


def check_dependent_modules(ctx, modules=[]):
    """Checks if any of the provided modules have module dependencies."""

    for module in modules:
        dependent_modules = ctx.modules.data.get(module, {}).get("dependentModules", [])
        if not dependent_modules:
            continue
        for dependent_module in dependent_modules:
            if not dependent_module in modules:
                modules.insert(0, dependent_module)
                ctx.logger.verbose(
                    f"Module dependency for module '{module}' will be included: '{dependent_module}'",
                )
    return list(set(modules))


def generate_identifier(identifiers=None):
    """Returns an 'object identifier' string used for creating log messages,
    e.g. '[ID: 12345] [Name: trino]'.

    ### Parameters
    - `identifiers`: Dictionary of "identifier_value": "identifier_key" pairs.

    ### Usage
    ```python
    identifier = generate_identifier(
        {"ID": container.short_id, "Name": container.name}
    ) # Will Spit out -> "[ID: 12345] [Name: trino]"
    ```"""

    if not identifiers:
        raise handle_missing_param(list(locals().keys()))

    identifier = []
    for key, value in identifiers.items():
        identifier.append(f"[{key}: {value}]")
    return " ".join(identifier)


def parse_key_value_pair(
    key_value_pair, err_type=err.MinitrinoError, key_to_upper=True
):
    """Parses a key-value pair in string form and returns the resulting pair as
    both a 2-element list. If the string cannot be split by "=", a
    MinitrinoError is raised.

    ### Parameters
    - `key_value_pair`: A string formatted as a key-value pair, i.e.
      `"STARBURST_VER=388-e"`.
    - `err_type`: The exception to raise if an "=" delimiter is not in the
      key-value pair. Defaults to `MinitrinoError`.
    - `key_to_upper`: If `True`, the key will be forced to uppercase.

    ### Return Values
    - A list `[k, v]`, but will return `None` if the stripped input is an empty
      string."""

    # Return None of empty string or special char (i.e. '\n')
    key_value_pair = key_value_pair.strip()
    if not key_value_pair:
        return None

    key_value_pair = key_value_pair.split("=", 1)
    err_msg = (
        f"Invalid key-value pair: '{'='.join(key_value_pair)}'. "
        f"Key-value pairs should be formatted as 'KEY=VALUE'"
    )

    # Raise an error if the key has no value
    if not key_value_pair[0]:
        raise err_type(err_msg)

    if isinstance(key_value_pair, list):
        for i in range(len(key_value_pair)):
            key_value_pair[i] = key_value_pair[i].strip()
        if not key_value_pair[0]:
            raise err_type(err_msg)
        elif key_to_upper:
            key_value_pair[0] = key_value_pair[0].upper()
    if not len(key_value_pair) == 2:
        raise err_type(err_msg)

    return key_value_pair


def get_cli_ver():
    """Returns the version of the Minitrino CLI."""

    return pkg_resources.require("Minitrino")[0].version


def get_lib_ver(library_path=""):
    """Returns the version of the Minitrino library.

    ### Parameters
    - `library_path`: The Minitrino library directory."""

    version_file = os.path.join(library_path, "version")
    try:
        with open(version_file, "r") as f:
            for line in f:
                line = line.strip()
                if line:
                    return line
            return "NOT INSTALLED"
    except:
        return "NOT INSTALLED"


def validate_yes(response=""):
    """Validates 'yes' user input. Returns `True` if a 'yes' input is
    detected."""

    response = response.replace(" ", "")
    if response.lower() == "y" or response.lower() == "yes":
        return True
    return False



#!usr/bin/env/python3
# -*- coding: utf-8 -*-

import os
import click

from minitrino import components

pass_environment = click.make_pass_decorator(components.Environment, ensure=True)


class CommandLineInterface(click.MultiCommand):
    def list_commands(self, ctx):
        cmd_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "cmd"))
        retval = []
        for filename in os.listdir(cmd_dir):
            if filename.endswith(".py") and filename.startswith("cmd_"):
                retval.append(filename[4:-3].replace("_", "-"))
        retval.sort()
        return retval

    def get_command(self, ctx, name):
        try:
            mod = __import__(
                f"minitrino.cmd.cmd_{name.replace('-', '_')}", None, None, ["cli"]
            )
        except ImportError:
            return
        return mod.cli


@click.command(cls=CommandLineInterface)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help=("""Enable verbose output."""),
)
@click.option(
    "-e",
    "--env",
    default=[],
    type=str,
    multiple=True,
    help=(
        """Add or override environment variables. 

        To learn more about environment variables and the order of precedence,
        visit the project wiki at:
        https://github.com/jefflester/minitrino/wiki/Environment-Variables-and-Config"""
    ),
)
@pass_environment
def cli(ctx, verbose, env):
    """Welcome to the Minitrino command line interface.

    To report issues and ask questions, please file a GitHub issue and apply a
    descriptive label at the GitHub repository:
    https://github.com/jefflester/minitrino
    """

    ctx._user_init(verbose, env)



#!usr/bin/env/python3
# -*- coding: utf-8 -*-

import os
import sys
import click
import shutil

from minitrino.cli import pass_environment
from minitrino import errors as err
from minitrino import utils


@click.command(
    "lib-install",
    help="Install the Minitrino library.",
)
@click.option(
    "-v",
    "--version",
    default="",
    type=str,
    help="Library version.",
)
@utils.exception_handler
@pass_environment
def cli(ctx, version):
    """Library installation command for Minitrino."""

    if not version:
        version = utils.get_cli_ver()

    lib_dir = os.path.join(ctx.minitrino_user_dir, "lib")
    if os.path.isdir(lib_dir):
        response = ctx.logger.prompt_msg(
            f"The Minitrino library at {lib_dir} will be overwritten. "
            f"Continue? [Y/N]"
        )
        if utils.validate_yes(response):
            ctx.logger.verbose("Removing existing library directory...")
            shutil.rmtree(lib_dir)
        else:
            ctx.logger.info("Opted to skip library installation.")
            sys.exit(0)

    download_and_extract(version)
    ctx.logger.info("Library installation complete.")


@pass_environment
def download_and_extract(ctx, version=""):
    github_uri = (
        f"https://github.com/jefflester/minitrino/archive/refs/tags/{version}.tar.gz"
    )
    tarball = os.path.join(ctx.minitrino_user_dir, f"{version}.tar.gz")
    file_basename = f"minitrino-{version}"  # filename after unpacking
    lib_dir = os.path.join(ctx.minitrino_user_dir, file_basename, "src", "lib")

    try:
        # Download the release tarball
        cmd = f"curl -fsSL {github_uri} > {tarball}"
        ctx.cmd_executor.execute_commands(cmd)
        if not os.path.isfile(tarball):
            raise err.MinitrinoError(
                f"Failed to download Minitrino library ({tarball} not found)."
            )

        # Unpack tarball and copy lib
        ctx.logger.verbose(
            f"Unpacking tarball at {tarball} and copying library...",
        )
        ctx.cmd_executor.execute_commands(
            f"tar -xzvf {tarball} -C {ctx.minitrino_user_dir}",
            f"mv {lib_dir} {ctx.minitrino_user_dir}",
        )

        # Check that the library is present
        lib_dir = os.path.join(ctx.minitrino_user_dir, "lib")
        if not os.path.isdir(lib_dir):
            raise err.MinitrinoError(
                f"Library failed to install (not found at {lib_dir})"
            )

        # Cleanup
        cleanup(tarball, file_basename)

    except Exception as e:
        cleanup(tarball, file_basename, False)
        raise err.MinitrinoError(str(e))


@pass_environment
def cleanup(ctx, tarball="", file_basename="", trigger_error=True):
    ctx.cmd_executor.execute_commands(
        f"rm -rf {tarball} {os.path.join(ctx.minitrino_user_dir, file_basename)}",
        trigger_error=trigger_error,
    )



#!usr/bin/env/python3
# -*- coding: utf-8 -*-

import click
import json

from minitrino.cli import pass_environment
from minitrino import errors as err
from minitrino import utils


@click.command(
    "modules",
    help=("""Display module metadata."""),
)
@click.option(
    "-m",
    "--module",
    "modules",
    default=[],
    type=str,
    multiple=True,
    help=("""A specific module to display metadata for."""),
)
@click.option(
    "-j",
    "--json",
    "json_format",
    is_flag=True,
    default=False,
    help=(
        """Print the resulting metadata in JSON form (shows additional module
        metadata)."""
    ),
)
@click.option(
    "-r",
    "--running",
    is_flag=True,
    default=False,
    help=("""Print metadata for all running modules."""),
)
@utils.exception_handler
@pass_environment
def cli(ctx, modules, json_format, running):
    """Module metadata command for Minitrino."""

    utils.check_lib(ctx)

    ctx.logger.info("Printing module metadata...")

    if not modules and not running:
        for module, module_dict in ctx.modules.data.items():
            log_info(module, module_dict, json_format)
        return

    if running:
        modules = ctx.modules.get_running_modules()

    for module in modules:
        module_dict = ctx.modules.data.get(module, {})
        if not module_dict:
            raise err.UserError(
                f"Invalid module: '{module}'",
                "Ensure the module you're referencing is in the Minitrino library.",
            )
        log_info(module, module_dict, json_format)


@pass_environment
def log_info(ctx, module_name="", module_dict={}, json_format=False):
    """Logs module metadata to the user's terminal."""

    if json_format:
        module_dict = {module_name: module_dict}
        ctx.logger.info(json.dumps(module_dict, indent=2))
    else:
        log_msg = [f"Module: {module_name}\n"]
        keys = ["description", "incompatibleModules", "dependentModules", "enterprise"]
        for key in keys:
            val = module_dict.get(key, None)
            if val is not None:
                key = list(key)
                key[0] = key[0].title()
                key = "".join(key)
                log_msg.extend(f"{key}: {val}\n")

        ctx.logger.info("".join(log_msg))



#!usr/bin/env/python3
# -*- coding: utf-8 -*-

import os
import click

from minitrino.cli import pass_environment
from minitrino import utils
from minitrino.settings import CONFIG_TEMPLATE

from shutil import rmtree


@click.command(
    "config",
    help=("""Edit the Minitrino user configuration file."""),
)
@click.option(
    "-r",
    "--reset",
    is_flag=True,
    default=False,
    help=(
        """Reset the Minitrino user configuration file and create a new config
        file from a template.

        WARNING: This will remove your configuration file (if it exists) and
        replace it with a template."""
    ),
)
@utils.exception_handler
@pass_environment
def cli(ctx, reset):
    """Config command for Minitrino."""

    if not os.path.isdir(ctx.minitrino_user_dir):
        ctx.logger.info(f"No {ctx.minitrino_user_dir} directory found. Creating...")
        os.mkdir(ctx.minitrino_user_dir)

    if os.path.isfile(ctx.config_file) and not reset:
        ctx.logger.verbose(
            f"Opening existing config file at path: {ctx.config_file}",
        )
        edit_file()
    elif os.path.isfile(ctx.config_file) and reset:
        response = ctx.logger.prompt_msg(f"Configuration file exists. Overwrite? [Y/N]")
        if utils.validate_yes(response):
            write_template()
            edit_file()
        else:
            ctx.logger.info(f"Opted out of recreating {ctx.minitrino_user_dir} file.")
    else:
        ctx.logger.verbose(
            f"No config file found at path: {ctx.config_file}. "
            f"Creating template config file and opening for edits...",
        )
        write_template()
        edit_file()


@pass_environment
def write_template(ctx):
    """Writes configuration template."""

    with open(ctx.config_file, "w") as config_file:
        config_file.write(CONFIG_TEMPLATE.lstrip())

    editor = ctx.env.get("TEXT_EDITOR", None)
    if not editor:
        editor = None


@pass_environment
def edit_file(ctx):
    """Gets the editor from user configuration and passes to the Click edit
    function if the value is present."""

    editor = ctx.env.get("TEXT_EDITOR", None)
    if not editor:
        editor = None

    click.edit(
        filename=ctx.config_file,
        editor=editor,
    )



#!usr/bin/env/python3
# -*- coding: utf-8 -*-

import sys
import click
from concurrent.futures import ThreadPoolExecutor, as_completed

from minitrino.cli import pass_environment
from minitrino import utils
from minitrino.settings import RESOURCE_LABEL


@click.command(
    "down",
    help=(
        """Bring down running Minitrino containers. This command follows the
        behavior of `docker compose down` where containers are both stopped and
        removed."""
    ),
)
@click.option(
    "-k",
    "--keep",
    is_flag=True,
    default=False,
    help=(
        """Does not remove containers; instead, containers will only be
        stopped."""
    ),
)
@click.option(
    "--sig-kill",
    is_flag=True,
    default=False,
    help=("""Stop Minitrino containers without a grace period."""),
)
@utils.exception_handler
@pass_environment
def cli(ctx, sig_kill, keep):
    """Down command for Minitrino. Exits with a 0 status code if there are no
    running minitrino containers."""

    utils.check_daemon(ctx.docker_client)
    utils.check_lib(ctx)

    containers = ctx.docker_client.containers.list(
        filters={"label": RESOURCE_LABEL}, all=True
    )

    if len(containers) == 0:
        ctx.logger.info("No containers to bring down.")
        sys.exit(0)

    # Helper function to stop a container
    def stop_container(container):
        identifier = utils.generate_identifier(
            {"ID": container.short_id, "Name": container.name}
        )
        if container.status == "running":
            if sig_kill:
                container.kill()
            else:
                container.stop()
            ctx.logger.verbose(f"Stopped container: {identifier}")
        return container

    # Helper function to remove a container
    def remove_container(container):
        identifier = utils.generate_identifier(
            {"ID": container.short_id, "Name": container.name}
        )
        container.remove()
        ctx.logger.verbose(f"Removed container: {identifier}")

    # Stop containers in parallel
    with ThreadPoolExecutor() as executor:
        stop_futures = {
            executor.submit(stop_container, container): container
            for container in containers
        }
        for future in as_completed(stop_futures):
            container = stop_futures[future]
            try:
                future.result()
            except Exception as e:
                ctx.logger.error(
                    f"Error stopping container '{container.name}': {str(e)}"
                )

    # Remove containers in parallel if not keeping them
    if not keep:
        with ThreadPoolExecutor() as executor:
            remove_futures = {
                executor.submit(remove_container, container): container
                for container in containers
            }
            for future in as_completed(remove_futures):
                container = remove_futures[future]
                try:
                    future.result()
                except Exception as e:
                    ctx.logger.error(
                        f"Error removing container '{container.name}': {str(e)}"
                    )

    ctx.logger.info("Brought down all Minitrino containers.")



#!usr/bin/env/python3
# -*- coding: utf-8 -*-

import os
from pathlib import Path
from setuptools import setup

HERE = Path(os.path.abspath(__file__)).resolve().parents[2]
README = (HERE / "readme.md").read_text()

setup(
    name="minitrino",
    version="2.1.4",
    description="A command line tool that makes it easy to run modular Trino environments locally.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/jefflester/minitrino",
    author="Jeff Lester",
    author_email="jeff.lester.dev@gmail.com",
    license="Apache-2.0",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    keyword="trino, docker, minitrino",
    python_requires=">=3.8",
    packages=["minitrino", "minitrino.cmd"],
    include_package_data=True,
    install_requires=[
        "click==8.1.7",
        "colorama",
        "docker==7.1.0",
        "PyYAML",
    ],
    entry_points={"console_scripts": ["minitrino=minitrino.cli:cli"]},
)



#!usr/bin/env/python3
# -*- coding: utf-8 -*-

import click

from minitrino.cli import pass_environment
from minitrino import utils


@click.command(
    "version",
    help=("""Display Minitrino CLI and library versions."""),
)
@utils.exception_handler
@pass_environment
def cli(ctx):
    """Version command for Minitrino."""

    cli_version = utils.get_cli_ver()
    ctx.logger.info(f"Minitrino version: {cli_version}")

    try:
        lib_version = utils.get_lib_ver(ctx.minitrino_lib_dir)
        ctx.logger.info(f"Library version: {lib_version}")
    except:
        ctx.logger.info("Library version: NOT INSTALLED")
